import argparse
from abc import ABC, abstractmethod
from dataclasses import dataclass, field


@dataclass(init=False)
class BasePlugin(ABC):
    name: str = field(init=False)
    aliases: tuple = field(init=False, default=())
    help: str = field(init=False, default=None)

    def __init__(self, subparser):
        self._subparser = subparser
        self.parser = self.create_parser(self._parser)

    @property
    def _parser(self) -> argparse.ArgumentParser:
        kwargs = {
            "name": self.name,
            "aliases": self.aliases,
        }
        if self.help is not None:
            kwargs["help"] = self.help

        p = self._subparser.add_parser(**kwargs)
        p.set_defaults(owner=self)
        return p

    def can_handle(self, owner) -> bool:
        return self == owner

    @abstractmethod
    def create_parser(self, parser: argparse.ArgumentParser):
        ...

    @abstractmethod
    def handle(self, namespace: argparse.Namespace) -> int:
        ...


__all__ = ["BasePlugin"]



from .cmd import run_command, kill_proc_tree
from .logging_utils import get_logger, LOGGER_NAME
from .types import Comparable

__all__ = ["run_command", "kill_proc_tree", "get_logger", "LOGGER_NAME", "Comparable"]



from abc import ABCMeta, abstractmethod
from typing import TypeVar


class _Comparable(metaclass=ABCMeta):
    @abstractmethod
    def __lt__(self, other) -> bool:
        ...


Comparable = TypeVar("Comparable", bound=_Comparable)



"""Different utilities that create a logger module, available for the hole system"""
import logging
import os
from sys import stdout

# Default format to use when logging messages:
# <LEVEL NAME>:    <MESSAGE>
LOG_DEFAULT_FORMAT = "%(levelname)s:\t%(message)s"

# Logger name that can be used globally to obtain this logger
LOGGER_NAME = r"orcha-logger"


def get_logger() -> logging.Logger:
    """Generates (or returns) an existing logger from the system
    that should be used globally on this program.
    Returns:
        logging.Logger: the system logger
    """
    log = logging.getLogger(LOGGER_NAME)

    # as we have set no `basicConfig`, newly created loggers
    # will evaluate this to False. If it existed, then it will
    # have at least one handler
    if log.hasHandlers():
        return log

    formatter = logging.Formatter(LOG_DEFAULT_FORMAT)
    handler = logging.StreamHandler(stream=stdout)
    level = os.environ.get("LOG_LEVEL", "INFO")
    handler.setLevel(level)
    handler.setFormatter(formatter)

    log.addHandler(handler)
    log.setLevel(level)

    return log



import argparse

parser = argparse.ArgumentParser(description="Orcha command line utility for handling services")
parser.add_argument(
    "--listen-address",
    metavar="ADDRESS",
    type=str,
    default="127.0.0.1",
    help="Listen address of the service",
)
parser.add_argument(
    "--port",
    metavar="N",
    type=int,
    default=50000,
    help="Listen port of the service",
)
parser.add_argument(
    "--key",
    metavar="KEY",
    type=str,
    default=None,
    help="Authentication key used for verifying clients",
)
subparsers = parser.add_subparsers(
    description="Available commands installed from external plugins",
    required=True,
    metavar="Commands",
)



import subprocess
import psutil
import signal
import shlex
from typing import Callable, Any

from .logging_utils import get_logger

log = get_logger()


def run_command(
    cmd: str,
    on_start: Callable[[subprocess.Popen], Any] = None,
    on_output: Callable[[str], Any] = None,
    on_finish: Callable[[int], Any] = None,
    cwd=None,
):
    if on_start is None:
        on_start = lambda _: None

    if on_output is None:
        on_output = lambda _: None

    if on_finish is None:
        on_finish = lambda _: None

    command = shlex.split(cmd)
    log.debug("$ %s", cmd)
    log.debug("> %s", command)

    with subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        bufsize=1,
        cwd=cwd,
    ) as proc:
        on_start(proc)
        for line in proc.stdout:
            on_output(line)

        ret = proc.wait()
    on_finish(ret)


def kill_proc_tree(pid, including_parent=True):
    try:
        parent = psutil.Process(pid)
        for child in parent.children(recursive=True):
            child.send_signal(signal.SIGTERM)
        if including_parent:
            parent.send_signal(signal.SIGTERM)
    except psutil.NoSuchProcess:
        log.warning("error while trying to kill proccess with id %d", pid)



"""When working with an Orcha project, properties are expected to be stored here"""
from typing import Optional


# When defining a <SyncManager>, "listen_address" specifies where to listen to
listen_address: str = "127.0.0.1"

# When creating a <SyncManager>, the port to listen to
port: int = 50000

# When creating a <SyncManager>, the authentication key used by other processes
authkey: Optional[bytes] = None

# Extra properties that you may want to store when working with the project
extras = {}


__all__ = ["listen_address", "port", "authkey", "extras"]



from dataclasses import dataclass, field

# from typing import Union


@dataclass(order=True)
class Message:
    # priority: Union[int, float] = field()
    id: int = field(compare=False)
    extras: dict = field(compare=False)



from dataclasses import dataclass, field

from daemon import DaemonContext

from orcha.lib.manager import Manager


@dataclass(frozen=True)
class ServiceWrapper:
    manager: Manager = field(init=True, compare=False, hash=False)
    context: DaemonContext = field(init=True, compare=False, hash=False, default=None)


__all__ = ["ServiceWrapper"]



import multiprocessing
import random
import subprocess
from queue import Queue, PriorityQueue
from threading import Lock, Thread, Event
from time import sleep
from typing import Union, Dict, List, Optional

from orcha import properties
from orcha.interfaces.message import Message
from orcha.interfaces.petition import Petition, EmptyPetition
from orcha.utils.cmd import kill_proc_tree
from orcha.utils.logging_utils import get_logger


log = get_logger()


class Processor:
    __instance__ = None

    def __new__(cls, *args, **kwargs):
        if Processor.__instance__ is None:
            instance = object.__new__(cls)
            instance.__must_init__ = True
            Processor.__instance__ = instance
        return Processor.__instance__

    def __init__(
        self,
        queue: multiprocessing.Queue = None,
        finishq: multiprocessing.Queue = None,
        manager=None,
    ):
        if self.__must_init__:
            if not all((queue, finishq, manager)):
                raise ValueError("queue & manager objects cannot be empty during init")

            self.lock = Lock()
            self.queue = queue
            self.finishq = finishq
            self.manager = manager
            self.running = True

            self._internalq = PriorityQueue()
            self._signals = Queue()
            self._threads: List[Thread] = []
            self._petitions: Dict[int, int] = {}
            self._gc_event = Event()
            self._process_t = Thread(target=self._process)
            self._internal_t = Thread(target=self._internal_process)
            self._finished_t = Thread(target=self._signal_handler)
            self._signal_t = Thread(target=self._internal_signal_handler)
            self._gc_t = Thread(target=self._gc)
            self._process_t.start()
            self._internal_t.start()
            self._finished_t.start()
            self._signal_t.start()
            self._gc_t.start()
            self.__must_init__ = False

    @property
    def running(self) -> bool:
        return self._running

    @running.setter
    def running(self, v: bool):
        with self.lock:
            self._running = v

    def exists(self, m: Union[Message, int]) -> bool:
        return self.manager.is_running(m)

    def enqueue(self, m: Message):
        self.queue.put(m)

    def finish(self, m: Union[Message, int]):
        if isinstance(m, Message):
            m = m.id

        log.debug("received petition for finish message with ID %d", m)
        self.finishq.put(m)

    def _process(self):
        log.debug("fixing internal digest key")
        multiprocessing.current_process().authkey = properties.authkey

        while self.running:
            log.debug("waiting for message...")
            m = self.queue.get()
            if m is not None:
                log.debug('converting message "%s" into a petition', m)
                p: Optional[Petition] = self.manager.convert_to_petition(m)
                if p is not None:
                    log.debug("> %s", p)
                    if self.exists(p.id):
                        log.warning("received message (%s) already exists", p)
                        p.queue.put(f'message with ID "{p.id}" already exists\n')
                        p.queue.put(None)
                        continue
                else:
                    log.debug(f'message "%s" is invalid, skipping...', m)
                    continue
            else:
                p = EmptyPetition()
            self._internalq.put(p)

    def _internal_process(self):
        last_seen = None
        while self.running:
            log.debug("waiting for internal petition...")
            p: Petition = self._internalq.get()
            if not isinstance(p, EmptyPetition):
                if p.condition(p):
                    log.debug('petition "%s" satisfied condition', p)
                    launch_t = Thread(target=self._start, args=(p,))
                    launch_t.start()
                    self._threads.append(launch_t)
                else:
                    log.debug('petition "%s" did not satisfy the condition, re-adding to queue', p)
                    self._internalq.put(p)
                    if p.id == last_seen:
                        log.debug(
                            'petition "%s" was already enqueued! waiting some time before polling',
                            p,
                        )
                        sleep(random.uniform(0.1, 0.5))
                    last_seen = p.id
            else:
                log.debug("received empty petition")

    def _start(self, p: Petition):
        log.debug('launching petition "%s"', p)

        def assign_pid(proc: Union[subprocess.Popen, int]):
            pid = proc if isinstance(proc, int) else proc.pid
            log.debug('assigning pid to "%d"', pid)
            self._petitions[p.id] = pid
            self.manager.on_start(p)

        try:
            # self.manager.on_start(p)
            p.action(assign_pid, p)
        except Exception as e:
            log.warning(
                'unhandled exception while running petition "%s" -> "%s"', p, e, exc_info=True
            )
        finally:
            log.debug('petition "%s" finished, triggering callbacks', p)
            self._petitions.pop(p.id, None)
            self._gc_event.set()
            self.manager.on_finish(p)

    def _signal_handler(self):
        log.debug("fixing internal digest key")
        multiprocessing.current_process().authkey = properties.authkey

        while self.running:
            log.debug("waiting for finish message...")
            m = self.finishq.get()
            self._signals.put(m)

    def _internal_signal_handler(self):
        while self.running:
            log.debug("waiting for internal signal...")
            m = self._signals.get()
            if isinstance(m, Message):
                m = m.id

            if m is not None:
                log.debug('received signal petition for message with ID "%d"', m)
                if m not in self._petitions:
                    log.warning('message with ID "%d" not found or not running!', m)

                pid = self._petitions[m]
                # proc.send_signal(signal.SIGINT)
                kill_proc_tree(pid, including_parent=False)
                log.debug('sent signal to process "%d" and all of its children', pid)

    def _gc(self):
        while self.running:
            self._gc_event.wait()
            self._gc_event.clear()
            for thread in self._threads:
                if not thread.is_alive():
                    log.debug('pruning thread "%s"', thread)
                    self._threads.remove(thread)

    def shutdown(self):
        try:
            log.info("finishing processor")
            self.running = False
            self.queue.put(None)
            self.finishq.put(None)
            self._gc_event.set()

            log.info("waiting for pending processes...")
            self._process_t.join()
            self._internal_t.join()

            log.info("waiting for pending signals...")
            self._finished_t.join()
            self._signal_t.join()

            log.info("waiting for garbage collector...")
            self._gc_t.join()

            log.info("waiting for pending operations...")
            for thread in self._threads:
                thread.join()

            log.info("finished")
        except Exception as e:
            log.critical(f"unexpected error during shutdown! -> {e}", exc_info=True)


__all__ = ["Processor"]



from abc import ABC, abstractmethod
import multiprocessing
from multiprocessing.managers import SyncManager
from queue import Queue
from typing import Callable, Optional, Union

from orcha.exceptions import ManagerShutdownError
from orcha.interfaces.message import Message

from orcha.utils.logging_utils import get_logger
from orcha.interfaces import Petition
from orcha.processor import Processor
from orcha import properties

log = get_logger()
_queue = multiprocessing.Queue()
_finish_queue = multiprocessing.Queue()


class Manager(ABC):
    def __init__(
        self,
        listen_address: str = properties.listen_address,
        port: int = properties.port,
        auth_key: bytes = properties.authkey,
        create_processor: bool = True,
        queue: Queue = None,
        finish_queue: Queue = None,
        is_client: bool = False,
    ):
        self.manager = SyncManager(address=(listen_address, port), authkey=auth_key)
        self.create_processor = create_processor
        self.is_client = is_client
        self._set_lock = multiprocessing.Lock()
        self._enqueued_messages = set()
        self._shutdown = multiprocessing.Value("b", False)

        # clients don't need any processor
        if create_processor and not is_client:
            log.debug("creating processor for <%s>", self)
            queue = queue or _queue
            finish_queue = finish_queue or _finish_queue
            self.processor = Processor(queue, finish_queue, self)

        log.debug("manager created - running setup...")
        self.setup()

    @property
    def processor(self) -> Processor:
        if not self.create_processor or self.is_client:
            raise RuntimeError("this manager has no processor")

        return self._processor

    @processor.setter
    def processor(self, processor: Processor):
        if not self.create_processor or self.is_client:
            raise RuntimeError("this manager does not support processors")

        self._processor = processor

    def connect(self):
        log.debug("connecting to manager")
        self.manager.connect()

    def start(self):
        log.debug("starting manager")
        self.manager.start()

    def serve(self):
        server = self.manager.get_server()
        server.serve_forever()

    def shutdown(self):
        self._shutdown.value = True
        try:
            if self.create_processor and not self.is_client:
                log.debug("shutting down processor")
                self.processor.shutdown()

            log.debug("finishing manager")
            self.manager.shutdown()
            self.manager.join()

            log.debug("parent handler finished")
        except Exception as e:
            log.critical(f"unexpected error during shutdown! -> {e}", exc_info=True)

    def join(self):
        log.debug("waiting for manager...")
        self.manager.join()
        log.debug("manager joined")

    def register(self, name: str, func: Optional[Callable] = None, **kwargs):
        log.debug('registering callable "%s" with name "%s"', func, name)
        self.manager.register(name, func, **kwargs)

        def temp(*args, **kwds):
            return getattr(self.manager, name)(*args, **kwds)

        setattr(self, name, temp)

    def send(self, m: Message):
        """stub"""

    def finish(self, m: Union[Message, int]):
        """stub"""

    def _add_message(self, m: Message):
        if not self._shutdown.value:
            return self.processor.enqueue(m)

        log.debug("we're off - enqueue petition not accepted for message with ID %d", m.id)
        raise ManagerShutdownError("manager has been shutdown - no more petitions are accepted")

    def _finish_message(self, m: Union[Message, int]):
        if not self._shutdown.value:
            return self.processor.finish(m)

        log.debug(
            "we're off - finish petition not accepted for message with ID %d",
            m.id if isinstance(m, Message) else m,
        )
        raise ManagerShutdownError("manager has been shutdown - no more petitions are accepted")

    def setup(self):
        send_fn = None if self.is_client else self._add_message
        finish_fn = None if self.is_client else self._finish_message

        self.register("send", send_fn)
        self.register("finish", finish_fn)

    def is_running(self, x: Union[Message, Petition, int]):
        if isinstance(x, (Message, Petition)):
            x = x.id

        with self._set_lock:
            return x in self._enqueued_messages

    @property
    def running_processes(self) -> int:
        with self._set_lock:
            return len(self._enqueued_messages)

    @abstractmethod
    def convert_to_petition(self, m: Message) -> Optional[Petition]:
        raise RuntimeError("must override!")

    @abstractmethod
    def on_start(self, p: Petition):
        with self._set_lock:
            self._enqueued_messages.add(p.id)
        log.info("starts execution of petition %d", p.id)

    @abstractmethod
    def on_finish(self, p: Petition):
        with self._set_lock:
            self._enqueued_messages.remove(p.id)
        log.info("finishes execution of petition %d", p.id)


__all__ = ["Manager"]



from .exceptions import *
from .utils import *
from .properties import *
from .handlers import *
from .interfaces import *
from .lib import *
from .processor import *
from orcha.plugins.plugin import *



import logging
import signal
from pwd import getpwnam
from typing import Iterable

import daemon
import daemon.pidfile

from orcha.interfaces.service import ServiceWrapper
from orcha.lib.manager import Manager
from orcha.utils.logging_utils import get_logger

log = get_logger()


def register_service(
    manager: Manager,
    *,
    pidfile: str = None,
    fds: Iterable[int] = None,
    user: str = None,
    group: str = None,
    cwd: str = "/",
    stop_signal: int = signal.SIGTERM,
) -> ServiceWrapper:
    pid_file = daemon.pidfile.PIDLockFile(pidfile)
    preserved_fds = [
        handler.stream.fileno()
        for handler in log.handlers
        if isinstance(handler, logging.FileHandler) and handler.stream is not None
    ]
    if fds is not None:
        preserved_fds.extend(fds)

    uid = getpwnam(user) if user is not None else None
    gid = getpwnam(group) if group is not None else None

    return ServiceWrapper(
        manager,
        daemon.DaemonContext(
            working_directory=cwd,
            umask=0o022,
            pidfile=pid_file,
            files_preserve=preserved_fds,
            uid=uid,
            gid=gid,
            signal_map={stop_signal: lambda *_: manager.shutdown()},
        ),
    )


def start_service(service: ServiceWrapper):
    if service.context is not None:
        with service.context:
            return service.manager.serve()

    try:
        service.manager.start()
        service.manager.join()
    except KeyboardInterrupt:
        service.manager.shutdown()
        exit(0)


__all__ = ["register_service", "start_service"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['orcha',
 'orcha.bin',
 'orcha.exceptions',
 'orcha.handlers',
 'orcha.interfaces',
 'orcha.lib',
 'orcha.plugins',
 'orcha.utils']

package_data = \
{'': ['*']}

install_requires = \
['psutil>=5.8.0,<6.0.0', 'python-daemon>=2.3.0,<3.0.0']

setup_kwargs = {
    'name': 'orcha',
    'version': '0.1.1b0',
    'description': 'System handler and orchestrator of multiple environments',
    'long_description': None,
    'author': 'Javier Alonso',
    'author_email': 'jalonso@teldat.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9.0,<4.0.0',
}


setup(**setup_kwargs)



from .petition import ProcT, ActionCallbackT, Petition
from .service import ServiceWrapper
from .message import Message

__all__ = ["ProcT", "ActionCallbackT", "Petition", "ServiceWrapper", "Message"]



import importlib
import pkgutil
import sys
from typing import Type

from ..plugins import BasePlugin
from orcha.utils.logging_utils import get_logger
import orcha.plugins


log = get_logger()


def iter_namespace(ns_pkg):
    # Specifying the second argument (prefix) to iter_modules makes the
    # returned name an absolute name instead of a relative one. This allows
    # import_module to work without having to do additional modification to
    # the name.
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


def main():
    import argparse

    parser = argparse.ArgumentParser(description="Orcha command line utility for handling services")
    parser.add_argument(
        "--listen-address",
        metavar="ADDRESS",
        type=str,
        default="127.0.0.1",
        help="Listen address of the service",
    )
    parser.add_argument(
        "--port",
        metavar="N",
        type=int,
        default=50000,
        help="Listen port of the service",
    )
    parser.add_argument(
        "--key",
        metavar="KEY",
        type=str,
        default=None,
        help="Authentication key used for verifying clients",
    )
    subparsers = parser.add_subparsers(
        description="Available commands installed from external plugins",
        required=True,
        metavar="Commands",
    )

    discovered_plugins = {
        name: importlib.import_module(name)
        for finder, name, ispkg in iter_namespace(orcha.plugins)
    }

    plugins = set()

    for plugin, mod in discovered_plugins:
        pl: Type[BasePlugin] = getattr(mod, "plugin", None)
        if pl is None:
            log.warning(
                f'invalid plugin specified for "{plugin}". Is there a plugin export class defined in __init__?'
            )
            continue

        plugins.add(pl(subparsers))

    args = parser.parse_args()
    for plugin in plugins:
        if plugin.can_handle(args.owner):
            return plugin.handle(args)


if __name__ == "__main__":
    sys.exit(main())



from queue import Queue
import subprocess
from abc import ABC
from dataclasses import dataclass, field
from typing import Callable, Union, NoReturn, Type, TypeVar

ProcT = Union[subprocess.Popen, int]
ActionCallbackT = Callable[[ProcT], NoReturn]

P = TypeVar("P", bound="Petition")


@dataclass(order=True)
class Petition(ABC):
    priority: int = field(init=False)
    id: int = field(compare=False)
    queue: Queue = field(compare=False, repr=False)
    action: Callable[[ActionCallbackT, Type[P]], NoReturn] = field(compare=False, repr=False)
    condition: Callable[[Type[P]], bool] = field(compare=False, repr=False)


@dataclass(init=False)
class EmptyPetition(Petition):
    priority = float("inf")
    id = -1
    queue = None
    action = None
    condition = None

    def __init__(self):
        pass


__all__ = ["ProcT", "ActionCallbackT", "Petition", "EmptyPetition"]
